from . import Board
from . import List
from ..misc.constants import API_URL
from ..utils.request import do_request

class Card:
	def __init__(self, data, list=None, trello_instance=None):
		self.parent = list
		self.list = list

		self._data = data

		self.id = data["id"]
		self.name = data["name"]
		self.closed = data["closed"]
		self.date_last_activity = data.get("dateLastActivity")
		self.desc = self.description = data["desc"]
		self.desc_data = data["descData"]
		self.id_board = data["idBoard"]
		self.id_list = data["idList"]
		self.id_short = data["idShort"]
		self.url = data["url"]
		self.subscribed = data["subscribed"]
		self.short_url = data["shortUrl"]
		self.pos = self.position = data["pos"]
		self._labels = data["labels"]

		self.labels = [] # TODO
		self.comments = [] # TODO

		self.synced = False
		self.trello_instance = trello_instance or getattr(list, "trello_instance", None)

	async def sync(self, data=None):
		self.synced = False
		self.labels.clear()
		self.comments.clear()

		if not data:
			data = await do_request(
				"GET",
				f"{API_URL}/cards/{self.id}",
				key=self.trello_instance.key,
				token=self.trello_instance.token,
				loop=self.trello_instance.loop,
			)

		old_data = self._data
		self._data = data

		self.id = data["id"]
		self.name = data["name"]
		self.closed = data["closed"]
		self.date_last_activity = data.get("dateLastActivity")
		self.desc = self.description = data["desc"]
		self.desc_data = data["descData"]
		self.id_board = data["idBoard"]
		self.id_list = data["idList"]
		self.id_short = data["idShort"]
		self.url = data["url"]
		self.subscribed = data["subscribed"]
		self.short_url = data["shortUrl"]
		self.pos = self.position = data["pos"]
		self._labels = data["labels"]

		if self.parent:
			if not self.parent.synced:
				await self.parent.sync()

			if self.id_board != self.parent.id:
				# card changed lists

				if old_data in self.parent._cards:
					self.parent.cards.remove(self)
					self.parent._cards.remove(old_data)

					if hasattr(self.parent, "parent") and self.trello_instance:
						board = await self.trello_instance.get_board(lambda b: b.id == self.id_board)
						if not board:
							await self.trello_instance.sync()

						new_list = await board.get_list(lambda l: l.id == self.id_list)

						new_list.cards.append(self)
						new_list._cards.append(self._data)

		self.synced = True

	async def move_to(self, list, board=None):
		list_id = List.List.resolve_id(list) # pylint: disable=E1101

		params = {
			"idList": list_id,
			"idBoard": self.parent.parent.id
		}

		if board:
			params["idBoard"] = Board.resolve_id(board)

		await do_request(
			"PUT",
			f"{API_URL}/cards/{self.id}",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=params
		)

		await self.sync()

	async def delete(self):
		await do_request(
			"DELETE",
			f"{API_URL}/cards/{self.id}",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
		)
		if self in self.parent.cards:
			self.parent.cards.remove(self)

		if self._data in self.parent._cards:
			self.parent._cards.remove(self._data)

	async def edit(self, **kwargs):
		await do_request(
			"PUT",
			f"{API_URL}/cards/{self.id}",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=kwargs
		)

	async def archive(self):
		if self in self.parent.cards:
			self.parent.cards.remove(self)

		if self._data in self.parent._cards:
			self.parent._cards.remove(self._data)

		await self.edit(closed=True)

	async def restore(self):
		if self not in self.parent.cards:
			self.parent.cards.append(self)

		if self._data not in self.parent._cards:
			self.parent._cards.append(self._data)

		await self.edit(closed=False)

	async def add_comment(self, text):
		# TODO: return a Comment object
		await do_request(
			"POST",
			f"{API_URL}/cards/{self.id}/actions/comments",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params={"text": text}
		)

	# aliases
	new_comment = create_comment = add_comment

	def __str__(self):
		return f"Card: {self.name} ({self.id})"

	def __repr__(self):
		return str(self)

	def __eq__(self, other):
		return self.id == other.id



from . import Board
from .Card import Card
from ..misc.constants import API_URL
from ..utils.request import do_request


class List:
	def __init__(self, data, cards=None, board=None):
		if not cards:
			cards = data.get("cards")

		self.parent = board
		self.board = board

		self._data = data

		self.name = data.get("name")
		self.idBoard = data.get("idBoard")
		self.pos = data.get("pos")
		self.subscribed = data.get("subscribed")
		self.id = data["id"]

		self._cards = cards or []
		self.cards = []

		self.synced = False
		self.trello_instance = getattr(board, "trello_instance", None)

		self._add_cards()

	@staticmethod
	def resolve_id(list):
		if isinstance(list, List):
			return list.id
		elif isinstance(List, str):
			return list
		else:
			# TODO: raise TrelloRuntimeError? ..or TrelloUsageError
			pass
	"""
	@staticmethod
	async def parse_from(list):
		if isinstance(list, List):
			return list
		else:
			list_json = await do_request(
				"GET",
				f"{API_URL}/lists/{self.id}/cards",
				key=self.trello_instance.key,
				token=self.trello_instance.token,
				loop=self.trello_instance.loop,
			)
	"""


	def _add_cards(self):
		if self._cards:
			for _card in self._cards:
				card = Card(_card, self)
				self.cards.append(card)

	async def sync(self, data=None, cards=None):
		self.synced = False

		if not data:
			data = await do_request(
				"GET",
				f"{API_URL}/lists/{self.id}",
				key=self.trello_instance.key,
				token=self.trello_instance.token,
				loop=self.trello_instance.loop,
			)

		self.name = data.get("name")
		self.idBoard = data.get("idBoard")
		self.pos = data.get("pos")
		self.subscribed = data.get("subscribed")
		self.id = data["id"]
		self._data = data
		self.cards.clear()
		self._cards.clear()

		if cards:
			self._cards = cards
		else:
			self._cards = await do_request(
				"GET",
				f"{API_URL}/lists/{self.id}/cards",
				key=self.trello_instance.key,
				token=self.trello_instance.token,
				loop=self.trello_instance.loop,
			)

		self._add_cards()
		self.synced = True

	async def get_cards(self):
		if not self.synced:
			await self.sync()

		return list(self.cards)

	async def get_card(self, predicate):
		if not self.synced:
			await self.sync()

		for card in self.cards:
			if predicate(card):
				return card

	async def archive(self):
		if not self.synced:
			await self.sync()

		params = {"value": True}
		await do_request(
			"PUT",
			f"{API_URL}/lists/{self.id}/closed",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=params
		)

		self.parent.lists.remove(self)

	async def restore(self):
		if not self.synced:
			await self.sync()

		params = {"value": False}
		await do_request(
			"PUT",
			f"{API_URL}/lists/{self.id}/closed",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=params
		)

		if self not in self.parent.boards:
			self.parent.board.append(self)

	async def edit(self, **kwargs):
		await do_request(
			"PUT",
			f"{API_URL}/lists/{self.id}",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=kwargs
		)
		await self.sync()

	async def create_card(self, name=None, desc=None, **kwargs):
		kwargs["idList"] = self.id

		if name:
			kwargs["name"] = name
		if desc:
			kwargs["desc"] = desc

		card_json = await do_request(
			"POST",
			f"{API_URL}/cards/",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			session=self.trello_instance.session,
			params=kwargs
		)

		card = Card(card_json, self)
		self._cards.append(card_json)
		self.cards.append(card)

		return card

	new_card = add_card = create_card

	@staticmethod
	async def from_board(board, key=None, token=None):
		board_id = Board.resolve_id(board)
		# TODO: finish this


	def __str__(self):
		return f"List: {self.name} ({self.id})"

	def __repr__(self):
		return str(self)

	def __eq__(self, other):
		return self.id == other.id



import re
from .List import List
from ..misc.constants import API_URL
from ..utils.request import do_request

compiled_board_regex = re.compile(r"b/(.*)/")

class Board:
	def __init__(self, json, trello_instance=None):
		self.id = json["id"]
		self.name = json["name"]
		self.closed = json["closed"]
		self.id_organization = json["idOrganization"]
		self.pinned = json["pinned"]
		self.url = json["url"]
		self.short_url = json["shortUrl"]
		self.prefs = json["prefs"]
		self.label_names = json["labelNames"]
		self.starred = json.get("starred", False)
		self.limits = json.get("limits", [])
		self.memberships = json.get("memberships", [])
		self.lists = []

		self.synced = False

		self.parent = trello_instance
		self.trello_instance = trello_instance


	async def sync(self, json=None):
		self.synced = False
		self.lists.clear()

		if not json:
			json = await do_request(
				"GET",
				f"{API_URL}/boards/{self.id}",
				key=self.trello_instance.key,
				token=self.trello_instance.token,
				loop=self.trello_instance.loop,
			)

		self.id = json["id"]
		self.name = json["name"]
		self.closed = json["closed"]
		self.id_organization = json["idOrganization"]
		self.pinned = json["pinned"]
		self.url = json["url"]
		self.short_url = json["shortUrl"]
		self.prefs = json["prefs"]
		self.label_names = json["labelNames"]
		self.starred = json.get("starred", False)
		self.limits = json.get("limits", [])
		self.memberships = json.get("memberships", [])
		self.lists.clear()

		await self.load_lists()

		self.synced = True

	async def create_list(self, name, **kwargs):
		if not self.synced:
			await self.sync()

		kwargs["name"] = name
		kwargs["idBoard"] = self.id

		json = await do_request(
			"POST",
			f"{API_URL}/lists/",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=kwargs
		)

		new_list = List(json, [], self)

		await new_list.sync()

		self.lists.append(new_list)

		return new_list

	async def load_lists(self, **kwargs):
		params = kwargs or {"cards": "open"}
		json = await do_request(
			"GET",
			f"{API_URL}/boards/{self.id}/lists",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=params
		)

		for trello_list in json:
			new_list = List(trello_list, trello_list.get("cards"), self)

			if new_list not in self.lists:
				self.lists.append(new_list)


	@staticmethod
	def resolve_id(board):
		board_id = None

		if isinstance(board, Board):
			return board.id
		elif isinstance(board, str):
			board_id = compiled_board_regex.search(board, re.IGNORECASE)

			if board_id:
				return board_id.group(1)

			return board
		else:
			# raise TrelloRuntimeError? ..or TrelloUsageError
			pass

	@staticmethod
	async def from_board(board, key=None, token=None, loop=None, trello_instance=None):
		board_id = Board.resolve_id(board)

		if trello_instance and (not key or not token):
			key = trello_instance.key
			token = trello_instance.token

		board = Board(await do_request(
			"GET",
			f"{API_URL}/boards/{board_id}",
			key=key,
			token=token,
			loop=loop,
		))

		if trello_instance:
			board.trello_instance = trello_instance
			board.parent = trello_instance

		await board.sync()

		return board

	async def delete(self):
		if not self.synced:
			await self.sync()

		await do_request(
			"DELETE",
			f"{API_URL}/boards/{self.id}",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
		)

		self.trello_instance.boards.remove(self)

	async def get_lists(self):
		if not self.synced:
			await self.sync()

		return list(self.lists)

	async def get_list(self, predicate):
		if not self.synced:
			await self.sync()

		for list in self.lists:
			if predicate(list):
				return list

	async def get_field(self, field):
		if not self.synced:
			await self.sync()

		if hasattr(self, field):
			return getattr(self, field)
		else:
			field_json = await do_request(
				"GET",
				f"{API_URL}/boards/{field}",
				key=self.trello_instance.key,
				token=self.trello_instance.token,
				loop=self.trello_instance.loop,
			)

			if field_json:
				setattr(self, field, field_json.get("_value"))

				return field_json.get("_value")

	async def edit(self, **kwargs):
		await do_request(
			"PUT",
			f"{API_URL}/boards/{self.id}",
			key=self.trello_instance.key,
			token=self.trello_instance.token,
			loop=self.trello_instance.loop,
			params=kwargs
		)

		await self.sync()

	def __str__(self):
		return f"Board: {self.name} ({self.id})"

	def __repr__(self):
		return str(self)

	def __eq__(self, other):
		return self.id == other.id



import asyncio
import aiohttp
from .misc.constants import API_URL
from .structures.Board import Board
from .utils.request import do_request


class Trello:
	def __init__(
			self,
			*,
			key=None,
			token=None,
			loop=asyncio.get_event_loop(),
			session=None,
			use_cache=True
		):
		"""Initializes a new Trello client.
		Parameters
		----------
		key: str [optional]
			The key used for Trello authentication.
		token: str [optional]
			The token used for Trello authentication.
		loop: asyncio loop [optional] [default=asyncio.get_event_loop()]
			An asyncio event loop used for HTTP requests.
		session: aiohttp.ClientSession [optional]
			aiohttp client session used for HTTP requests.
		use_cache: boolean [optional] [default=True]
			Whether to avoid HTTP requests on short intervals
			whenever possible. False = always make a new HTTP request.
			NOTE: this is currently not implemented.

		Note: if a key and token is not provided, you may only use the static methods of this package.
		"""

		self.key = key
		self.token = token
		self.loop = loop
		self.synced = False
		self.boards = []
		self._use_cache = use_cache

		if session:
			self.session = session
		else:
			self.session = aiohttp.ClientSession(loop=loop)

	async def get_board(self, a):
		if callable(a):
			for board in await self.get_boards():
				if a(board):
					return board
		else:
			board = await Board.from_board(a, trello_instance=self)

			return board


	async def get_boards(self):
		if not self.synced:
			await self.sync()

		return list(self.boards)


	async def sync(self):
		self.boards.clear()

		boards = await do_request(
			"GET",
			f"{API_URL}/members/me/boards",
			key=self.key,
			token=self.token,
			loop=self.loop,
			session=self.session
		)

		for raw_board in boards:
			board = Board(raw_board, self)

			if board not in self.boards:
				await board.sync()
				self.boards.append(board)

		if self._use_cache:
			self.synced = True

	async def create_board(self, name, **kwargs):
		kwargs["name"] = name
		board = Board(await do_request(
			"POST",
			f"{API_URL}/boards",
			key=self.key,
			token=self.token,
			loop=self.loop,
			params=kwargs,
			session=self.session
		), self)

		self.boards.append(board)
		return board

	new_board = create_board



class TrelloException(Exception):
	pass

class TrelloBadRequest(TrelloException):
	pass

class TrelloUnauthorized(TrelloException):
	pass

class TrelloHttpError(TrelloException):
	pass

class TrelloNotFound(TrelloException):
	pass

class TrelloJsonError(TrelloException):
	pass



from setuptools import setup, find_packages


setup(name="aiotrello",
	version="0.0.3",
	description="Async Trello library",
	url="https://github.com/bloxlink/aiotrello",
	keywords=["async", "trello"],
	author="tigerism",
	author_email="admin@blox.link",
	license="MIT",
	packages=find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3.5",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Operating System :: OS Independent"
	],
	long_description=open("README.md", "r").read(),
	long_description_content_type="text/markdown",
	zip_safe=False
)



import asyncio
import aiohttp
from .. import exceptions

async def do_request(method, url, params=None, key=None, token=None, loop=None, session=None):
	if not loop:
		loop = asyncio.get_event_loop()

	close_session = False
	if not session:
		session = aiohttp.ClientSession(loop=loop)
		close_session = True

	if key and token:
		url = f"{url}?key={key}&token={token}"

		if params:
			params = [f"&{x}={'true' if y is True else 'false' if y is False else y}" for x, y in params.items()]
			url = url + "".join(params)

	async with session.request(method, url) as response:
		if response.status == 400:
			raise exceptions.TrelloBadRequest(response)
		elif response.status == 401:
			raise exceptions.TrelloUnauthorized(response)
		elif response.status == 404:
			raise exceptions.TrelloNotFound(response)
		elif response.status != 200:
			raise exceptions.TrelloHttpError(response)

		json = await response.json()

		if close_session:
			await session.close()

		return json
